@@ -2,15 +2,18 @@ package main
 
 import (
 	"bufio"
+	"bytes"
 	"encoding/base64"
 	"fmt"
 	"io"
 	"io/ioutil"
 	"os"
 	"strings"
+	"text/template"
 
 	"time"
 
+	"github.com/Masterminds/sprig"
 	"github.com/Shopify/sarama"
 	"github.com/birdayz/kaf/pkg/partitioner"
 	pb "github.com/golang/protobuf/proto"
@@ -28,6 +31,8 @@ var (
 	bufferSizeFlag  int
 	inputModeFlag   string
 	avroSchemaID    int
+	avroKeySchemaID int
+	templateFlag    bool
 )
 
 func init() {
@@ -48,9 +53,13 @@ func init() {
 	produceCmd.Flags().Int32VarP(&partitionFlag, "partition", "p", -1, "Partition to produce to")
 
 	produceCmd.Flags().IntVarP(&avroSchemaID, "avro-schema-id", "", -1, "Value schema id for avro messsage encoding")
+	produceCmd.Flags().IntVarP(&avroKeySchemaID, "avro-key-schema-id", "", -1, "Key schema id for avro messsage encoding")
 
 	produceCmd.Flags().StringVarP(&inputModeFlag, "input-mode", "", "line", "Scanning input mode: [line|full]")
 	produceCmd.Flags().IntVarP(&bufferSizeFlag, "line-length-limit", "", 0, "line length limit in line input mode")
+
+	produceCmd.Flags().BoolVar(&templateFlag, "template", false, "run data through go template engine")
+
 }
 
 func readLines(reader io.Reader, out chan []byte) {
@@ -104,6 +113,13 @@ var produceCmd = &cobra.Command{
 			errorExit("Unable to create new sync producer: %v\n", err)
 		}
 
+		if avroSchemaID != -1 || avroKeySchemaID != -1 {
+			schemaCache = getSchemaCache()
+			if schemaCache == nil {
+				errorExit("Could not connect to schema registry")
+			}
+		}
+
 		out := make(chan []byte, 1)
 		switch inputModeFlag {
 		case "full":
@@ -139,13 +155,12 @@ var produceCmd = &cobra.Command{
 				errorExit("Failed to load key proto type")
 			}
 
-		}
-
-		if avroSchemaID != -1 {
-			schemaCache = getSchemaCache()
-			if schemaCache == nil {
-				errorExit("Error getting a instance of schemaCache")
+		} else if avroKeySchemaID != -1 {
+			avroKey, err := schemaCache.EncodeMessage(avroKeySchemaID, []byte(keyFlag))
+			if err != nil {
+				errorExit("Failed to encode avro key", err)
 			}
+			key = sarama.ByteEncoder(avroKey)
 		}
 
 		var headers []sarama.RecordHeader
@@ -179,7 +194,7 @@ var produceCmd = &cobra.Command{
 			} else if avroSchemaID != -1 {
 				avro, err := schemaCache.EncodeMessage(avroSchemaID, data)
 				if err != nil {
-					errorExit("Failed to encode avro", err)
+					errorExit("Failed to encode avro value", err)
 				}
 				data = avro
 			}
@@ -193,12 +208,34 @@ var produceCmd = &cobra.Command{
 			}
 
 			for i := 0; i < repeatFlag; i++ {
+
+				input := data
+
+				if templateFlag {
+					vars := map[string]interface{}{}
+					vars["i"] = i
+					tpl := template.New("kaf").Funcs(sprig.TxtFuncMap())
+
+					tpl, err = tpl.Parse(string(data))
+					if err != nil {
+						errorExit("failed to parse go template: %v", err)
+					}
+
+					buf := bytes.NewBuffer(nil)
+
+					if err := tpl.Execute(buf, vars); err != nil {
+						errorExit("failed to execute go template: %v", err)
+					}
+
+					input = buf.Bytes()
+				}
+
 				msg := &sarama.ProducerMessage{
 					Topic:     args[0],
 					Key:       key,
 					Headers:   headers,
 					Timestamp: ts,
-					Value:     sarama.ByteEncoder(data),
+					Value:     sarama.ByteEncoder(input),
 				}
 				if partitionFlag != -1 {
 					msg.Partition = partitionFlag
