@@ -1,6 +1,5 @@
 import dataclasses as dc
 import functools
-import hashlib
 import json
 import logging
 import random
@@ -11,6 +10,15 @@ import time
 from confluent_kafka import Consumer, Producer, KafkaException, KafkaError
 from retrying import retry
 
+
+def retry_if_buffer_error_or_retriable(exception):
+    """Return True if BufferError or retriable KafkaError, False otherwise"""
+    return (
+        isinstance(exception, KafkaException) and exception.args[0].retriable()
+        or
+        isinstance(exception, BufferError)
+    )
+
 class KafkaApp:
 
     def __init__(self, name, consumer_config, producer_config, consumer_batch_size=1, consumer_timeout=60):
@@ -23,7 +31,7 @@ class KafkaApp:
         self.consumer_batch_size = consumer_batch_size
         self.consumer_timeout = consumer_timeout
         self.on_processed_callbacks = []
-        # self.on_failed_callbacks = []
+        self.on_failed_callbacks = []
         # self.consumer = Consumer(consumer_config)
         # self.producer = Producer(producer_config)
         signal.signal(signal.SIGINT, self.exit_gracefully)
@@ -36,72 +44,47 @@ class KafkaApp:
         self.logger.info('Exiting gracefully')
         sys.exit()
 
-    @retry(wait_fixed=5000)
+    @retry(wait_fixed=5000, retry_on_exception=retry_if_buffer_error_or_retriable)
     def _initialise_clients(self):
         """
         Try to initialise until successful
         """
-        try:
-            self.logger.info('Initialising clients')
-            self.consumer = Consumer(self.consumer_config)
-            self.producer = Producer(self.producer_config)
-            topics = list(self.subs.keys())
-            self.logger.debug(f'Subscribing to topics: {topics}')
-            self.consumer.subscribe(topics)
-        except Exception as e:
-            self.logger.error(e)
-            raise e
+        self.logger.info('Trying to initialise clients...')
+        self.consumer = Consumer(self.consumer_config)
+        self.producer = Producer(self.producer_config)
+        topics = list(self.subs.keys())
+        self.logger.debug(f'Subscribing to topics: {topics}')
+        self.consumer.subscribe(topics)
+        self.logger.info('Clients initialised')
 
-
-    @retry(wait_fixed=5000)
+    @retry(wait_fixed=5000, retry_on_exception=retry_if_buffer_error_or_retriable)
     def _consume_messages(self):
         """
         Try to consume until successful (unless error is fatal)
         """
-        try:
-            self.logger.info('Consuming message')
-            msgs = self.consumer.consume(
-                num_messages=self.consumer_batch_size,
-                timeout=self.consumer_timeout
-            )
-            return msgs
-        except KafkaException as e:
-            kafka_error = e.args[0]
-            self.logger.error(kafka_error)
-            if not kafka_error.fatal(): raise e
+        return self.consumer.consume(
+            num_messages=self.consumer_batch_size,
+            timeout=self.consumer_timeout
+        )
 
-
-    @retry(wait_fixed=5000)
+    @retry(wait_fixed=5000, retry_on_exception=retry_if_buffer_error_or_retriable)
     def _produce_message(self, key, value, publish_to):
         """
         Try to produce until successful (unless error is fatal)
         """
-        try:
-            self.logger.debug(f'Producing message: KEY={key}, VALUE={value}')
-            self.logger.info(f'Producing message')
-            self.producer.produce(
-                key=key,
-                value=value,
-                topic=publish_to
-            )
-            self.producer.poll(0)
-        except KafkaException as e:
-            kafka_error = e.args[0]
-            self.logger.error(kafka_error)
-            if not kafka_error.fatal(): raise e
+        self.producer.produce(
+            key=key,
+            value=value,
+            topic=publish_to
+        )
+        self.producer.poll(0)
 
-
-    @retry(wait_fixed=5000)
+    @retry(wait_fixed=5000, retry_on_exception=retry_if_buffer_error_or_retriable)
     def _commit_message(self, msg):
         """
         Try to commit until successful (unless error is fatal)
         """
-        try:
-            self.consumer.commit(msg)
-        except KafkaException as e:
-            kafka_error = e.args[0]
-            self.logger.error(kafka_error)
-            if not kafka_error.fatal(): raise e
+        self.consumer.commit(msg)
 
     def run(self):
         """
@@ -118,7 +101,6 @@ class KafkaApp:
 
         """
         self.logger.debug('Run loop started')
-        # Try to initialise clients until successful
         self._initialise_clients()
 
         # Loop forever
@@ -128,12 +110,13 @@ class KafkaApp:
             self.logger.debug('Iteration started')
 
             # Try to consume messages until successful
+            self.logger.info('Consuming messages')
             msgs = self._consume_messages()
             if len(msgs) == 0:
                 self.logger.info(f'No messages consumed for {self.consumer_timeout} seconds')
             else:
                 self.logger.info(f'Consumed {len(msgs)} message(s)')
-            for msg in msgs:
+            for i, msg in enumerate(msgs):
                 # Case 1a: msg has retriable error => don't commit
                 # Case 1b: msg has fatal error => commit
                 # Case 2: msg was processed successfully => commit
@@ -141,65 +124,57 @@ class KafkaApp:
 
                     # Completely process each message before continuing to next
                     try:
+                        i += 1
                         t0 = time.perf_counter()
-                        commit = False
-                        sha256_value = self._get_sha256_hash(msg.value())
-                        self.logger.info(f'Processing message; SHA256_VALUE={sha256_value}')
-                        self.logger.debug(f'Processing message; VALUE={msg.value()}')
+                        self.logger.info(f'Input message[{i}] processing started')
 
-                        if msg.error() is not None:
+                        error = msg.error()
+                        if error is not None:
                             # Case 1a / 1b
-                            commit = not msg.error().retriable()
                             if error.code() == KafkaError._PARTITION_EOF:
                                 self.logger.info(
                                     f' {msg.topic()}[{msg.partition()}] reached end \
                                     of offset {msg.offset()}'
                                 )
                             else:
-                                self.logger.error(msg.error())
+                                self.logger.error(error)
                         else:
                             #
                             process_output = self._process_message(msg)
                             # Publish results
-                            for i, (value, key, publish_to) in enumerate(process_output):
-                                sha256_prod = self._get_sha256_hash(value)
+                            for j, (value, key, publish_to) in enumerate(process_output):
+                                j += 1
                                 self._produce_message(
                                     key=key,
                                     value=value,
                                     publish_to=publish_to
                                 )
-                                self.logger.info(f'Produced message[{i+1}]; SHA256_VALUE={sha256_prod}')
-                            self.logger.info(f'Processed message; SHA256_VALUE={sha256_value}')
+                                self.logger.info(f'Output message[{j}] produced')
                             # We don't care if callback raises an Exception
                             t1 = time.perf_counter()
-                            commit = True
                             for callback in self.on_processed_callbacks:
-                                callback(msg, t1 - t0)
-                    except KafkaException as e:
-                        kafka_error = e.args[0]
-                        commit = not kafka_error.retriable()
-                        self.logger.exception(e)
+                                try:
+                                    callback(msg, t1 - t0)
+                                except Exception as e_inner:
+                                    self.logger.exception(e)
                     except Exception as e:
-                        self.logger.exception(e)
+                        try:
+                            self.logger.exception(e)
+                            for callback in self.on_failed_callbacks:
+                                callback(msg, e)
+                        except Exception as e_inner:
+                            self.logger.exception(e)
                     finally:
-                        if commit:
+                        try:
                             self._commit_message(msg)
-                            self.logger.info(f'Committed message; SHA256_VALUE={sha256_value}')
-                        else:
-                            self.logger.info(f'Will not commit message; SHA256_VALUE={sha256_value}')
+                            self.logger.info(f'Input message[{i}] committed')
+                        except Exception as e:
+                            self.logger.exception(e)
+                            self.logger.error(f'Input message[{i}] not committed')
 
             iter_t1 = time.perf_counter()
             self.logger.debug(f'Iteration completed in {iter_t1 - iter_t0} seconds')
 
-    def _get_sha256_hash(self, value):
-        """
-        Returns the SHA256 hash of the input value (if not None), otherwise returns '-'.
-        """
-        if value is not None:
-            return hashlib.sha256(value).hexdigest()
-        else:
-            return '-'
-
     def _process_message(self, msg):
         """
         Process a single message
@@ -209,7 +184,7 @@ class KafkaApp:
         input_bytes = msg.value()
         self.logger.debug(f'Found {len(subs)} function(s) subscribed to topic "{topic}"')
         for func, publish_to, accepts, returns in subs:
-            self.logger.info(f'Calling user function "{func.__name__}"')
+            self.logger.debug(f'Calling user function "{func.__name__}"')
             input_obj = self._parse(input_bytes, accepts)
             for output_obj, key in func(input_obj):
                 if publish_to is None: continue
@@ -282,6 +257,6 @@ class KafkaApp:
         return func
 
 
-    # def on_failed(self, func):
-    #     self.on_failed_callbacks.append(func)
-    #     return func
+    def on_failure(self, func):
+        self.on_failed_callbacks.append(func)
+        return func
